@@ -1,20 +1,22 @@
-/***************************************************************************
-                          mymoneystoragesql.h
-                          -------------------
-    begin                : 11 November 2005
-    copyright            : (C) 2005 by Tony Bloomfield
-    email                : tonybloom@users.sourceforge.net
-                         : Fernando Vilas <fvilas@iname.com>
- ***************************************************************************/
-
-/***************************************************************************
- *                                                                         *
- *   This program is free software; you can redistribute it and/or modify  *
- *   it under the terms of the GNU General Public License as published by  *
- *   the Free Software Foundation; either version 2 of the License, or     *
- *   (at your option) any later version.                                   *
- *                                                                         *
- ***************************************************************************/
+/*
+ * This file is part of KMyMoney, A Personal Finance Manager for KDE
+ * Copyright (C) 2005 Tony Bloomfield <tonybloom@users.sourceforge.net>
+ * Copyright (C)      Fernando Vilas <fvilas@iname.com>
+ * Copyright (C) 2014 Christian Dávid <christian-david@web.de>
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License
+ * as published by the Free Software Foundation; either version 2
+ * of the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
 
 #ifndef MYMONEYSTORAGESQL_H
 #define MYMONEYSTORAGESQL_H
@@ -26,6 +28,7 @@
 #include <QSqlError>
 #include <QList>
 #include <QStack>
+#include <QSet>
 
 #include <QtDebug>
 
@@ -53,6 +56,7 @@ class QIODevice;
 #include "mymoneytransactionfilter.h"
 #include "mymoneydbdef.h"
 #include "mymoneydbdriver.h"
+#include "databasestoreableobject.h"
 
 class MyMoneyDbDriver;
 
@@ -114,6 +118,8 @@ class IMyMoneySerialize;
 /**
   * The MyMoneyDbColumn class is a base type for generic db columns.
   * Derived types exist for several common column types.
+  *
+  * @todo Remove unneeded columns which store the row count of tables from kmmFileInfo
   */
 class MyMoneyStorageSql : public IMyMoneyStorageFormat, public QSqlDatabase, public KShared
 {
@@ -156,30 +162,39 @@ public:
    * @return void
    *
    */
-  bool readFile(void);
+  bool readFile();
   /**
    * MyMoneyStorageSql write/update the database from storage
    *
    * @return void
    *
    */
-  bool writeFile(void);
+  bool writeFile();
 
   /**
-  * MyMoneyStorageSql generalized error routine
-  *
-  * @return : error message to be displayed
-  *
+   * MyMoneyStorageSql generalized error routine
+   *
+   * @return : error message to be displayed
+   *
    */
   const QString& lastError() const {
     return (m_error);
   };
+
+  /**
+   * MyMoneyStorageSql get highest ID number from the database
+   *
+   * @return : highest ID number
+   */
+  long unsigned highestNumberFromIdString(QString tableName, QString tableField, int prefixLength);
+
   /**
    * This method is used when a database file is open, and the data is to
    * be saved in a different file or format. It will ensure that all data
    * from the database is available in memory to enable it to be written.
    */
   virtual void fillStorage();
+
   /**
     * The following functions correspond to the identically named (usually) functions
     * within the Storage Manager, and are called to update the database
@@ -189,7 +204,7 @@ public:
   void modifyInstitution(const MyMoneyInstitution& inst);
   void removeInstitution(const MyMoneyInstitution& inst);
   void addPayee(const MyMoneyPayee& payee);
-  void modifyPayee(const MyMoneyPayee& payee);
+  void modifyPayee(MyMoneyPayee payee);
   void removePayee(const MyMoneyPayee& payee);
   void addTag(const MyMoneyTag& tag);
   void modifyTag(const MyMoneyTag& tag);
@@ -217,6 +232,12 @@ public:
   void addBudget(const MyMoneyBudget& bud);
   void modifyBudget(const MyMoneyBudget& bud);
   void removeBudget(const MyMoneyBudget& bud);
+  void addOnlineJob(const onlineJob& job);
+  void modifyOnlineJob(const onlineJob& job);
+  void removeOnlineJob(const onlineJob& job);
+  void addPayeeIdentifier(payeeIdentifier& ident);
+  void modifyPayeeIdentifier(const payeeIdentifier& ident);
+  void removePayeeIdentifier(const payeeIdentifier& ident);
 
   unsigned long transactionCount(const QString& aid = QString()) const;
   inline const QHash<QString, unsigned long> transactionCountMap() const {
@@ -241,13 +262,16 @@ public:
   const QMap<QString, MyMoneyInstitution> fetchInstitutions(const QStringList& idList = QStringList(), bool forUpdate = false) const;
   const QMap<QString, MyMoneyPayee> fetchPayees(const QStringList& idList = QStringList(), bool forUpdate = false) const;
   const QMap<QString, MyMoneyTag> fetchTags(const QStringList& idList = QStringList(), bool forUpdate = false) const;
+  const QMap<QString, onlineJob> fetchOnlineJobs(const QStringList& idList = QStringList(), bool forUpdate = false) const;
   const MyMoneyPriceList fetchPrices(const QStringList& fromIdList = QStringList(), const QStringList& toIdList = QStringList(), bool forUpdate = false) const;
-  const MyMoneyPrice fetchSinglePrice(const QString& fromIdList, const QString& toIdList, const QDate& date, bool exactDate, bool forUpdate = false) const;
+  MyMoneyPrice fetchSinglePrice(const QString& fromId, const QString& toId, const QDate& date_, bool exactDate, bool = false) const;
   const QMap<QString, MyMoneyReport> fetchReports(const QStringList& idList = QStringList(), bool forUpdate = false) const;
   const QMap<QString, MyMoneySchedule> fetchSchedules(const QStringList& idList = QStringList(), bool forUpdate = false) const;
   const QMap<QString, MyMoneySecurity> fetchSecurities(const QStringList& idList = QStringList(), bool forUpdate = false) const;
   const QMap<QString, MyMoneyTransaction> fetchTransactions(const QString& tidList = QString(), const QString& dateClause = QString(), bool forUpdate = false) const;
   const QMap<QString, MyMoneyTransaction> fetchTransactions(const MyMoneyTransactionFilter& filter) const;
+  payeeIdentifier fetchPayeeIdentifier(const QString& id) const;
+  const QMap<QString, payeeIdentifier> fetchPayeeIdentifiers(const QStringList& idList = QStringList()) const;
   bool isReferencedByTransaction(const QString& id) const;
 
   void readPayees(const QString&);
@@ -275,6 +299,8 @@ public:
   long unsigned getNextInstitutionId() const;
   long unsigned getNextPayeeId() const;
   long unsigned getNextTagId() const;
+  long unsigned getNextOnlineJobId() const;
+  long unsigned getNextPayeeIdentifierId() const;
   long unsigned getNextReportId() const;
   long unsigned getNextScheduleId() const;
   long unsigned getNextSecurityId() const;
@@ -289,6 +315,8 @@ public:
   long unsigned incrementScheduleId();
   long unsigned incrementSecurityId();
   long unsigned incrementTransactionId();
+  long unsigned incrementOnlineJobId();
+  long unsigned incrementPayeeIdentfierId();
 
   void loadAccountId(const unsigned long& id);
   void loadTransactionId(const unsigned long& id);
@@ -299,6 +327,8 @@ public:
   void loadSecurityId(const unsigned long& id);
   void loadReportId(const unsigned long& id);
   void loadBudgetId(const unsigned long& id);
+  void loadOnlineJobId(const unsigned long& id);
+  void loadPayeeIdentifierId(const unsigned long& id);
 
   /**
     * This method allows to modify the precision with which prices
@@ -313,27 +343,42 @@ public:
   static void setStartDate(const QDate &startDate);
 
 private:
-  //void init(void);
   bool fileExists(const QString& dbName);
-  // a function to build a comprehensive error message
+  /** @brief a function to build a comprehensive error message for an SQL error */
   QString& buildError(const QSqlQuery& q, const QString& function, const QString& message) const;
+
+  /** @copydoc buildError */
   QString& buildError(const QSqlQuery& q, const QString& function, const QString& message,
                       const QSqlDatabase*) const;
-  // write routines
-  void writeUserInformation(void);
-  void writeInstitutions(void);
-  void writePayees(void);
-  void writeTags(void);
-  void writeAccounts(void);
-  void writeTransactions(void);
-  void writeSchedules(void);
-  void writeSecurities(void);
-  void writePrices(void);
-  void writeCurrencies(void);
-  void writeFileInfo(void);
-  void writeReports(void);
-  void writeBudgets(void);
 
+  /**
+   * @name writeFromStorageMethods
+   * @{
+   * These method write all data from m_storage to the database. Data which is
+   * stored in the database is deleted.
+   */
+  void writeUserInformation();
+  void writeInstitutions();
+  void writePayees();
+  void writeTags();
+  void writeAccounts();
+  void writeTransactions();
+  void writeSchedules();
+  void writeSecurities();
+  void writePrices();
+  void writeCurrencies();
+  void writeFileInfo();
+  void writeReports();
+  void writeBudgets();
+  void writeOnlineJobs();
+  /** @} */
+
+  /**
+   * @name writeMethods
+   * @{
+   * These methods bind the data fields of MyMoneyObjects to a given query and execute the query.
+   * This is helpfull as the query has usually an update and a insert format.
+   */
   void writeInstitutionList(const QList<MyMoneyInstitution>& iList, QSqlQuery& q);
   void writePayee(const MyMoneyPayee& p, QSqlQuery& q, bool isUserInfo = false);
   void writeTag(const MyMoneyTag& p, QSqlQuery& q);
@@ -350,23 +395,33 @@ private:
   void writeReport(const MyMoneyReport& rep, QSqlQuery& q);
   void writeBudget(const MyMoneyBudget& bud, QSqlQuery& q);
   void writeKeyValuePairs(const QString& kvpType, const QVariantList& kvpId, const QList<QMap<QString, QString> >& pairs);
+  void writeOnlineJob(const onlineJob& job, QSqlQuery& query);
+  void writePayeeIdentifier(const payeeIdentifier& pid, QSqlQuery& query);
+  /** @} */
 
-  // read routines
-  void readFileInfo(void);
-  void readLogonData(void);
-  void readUserInformation(void);
-  void readInstitutions(void);
-  void readAccounts(void);
+  /**
+   * @name readMethods
+   * @{
+   */
+  void readFileInfo();
+  void readLogonData();
+  void readUserInformation();
+  void readInstitutions();
+  void readAccounts();
   void readTransactions(const QString& tidList = QString(), const QString& dateClause = QString());
   void readSplit(MyMoneySplit& s, const QSqlQuery& q) const;
   const MyMoneyKeyValueContainer readKeyValuePairs(const QString& kvpType, const QString& kvpId) const;
   const QHash<QString, MyMoneyKeyValueContainer> readKeyValuePairs(const QString& kvpType, const QStringList& kvpIdList) const;
-  void readSchedules(void);
-  void readSecurities(void);
-  void readPrices(void);
-  void readCurrencies(void);
-  void readReports(void);
-  void readBudgets(void);
+  void readSchedules();
+  void readSecurities();
+  void readPrices();
+  void readCurrencies();
+  void readReports();
+  void readBudgets();
+  /** @} */
+
+  template<long unsigned MyMoneyStorageSql::* cache>
+  long unsigned int getNextId(const QString& table, const QString& id, const int prefixLength) const;
 
   void deleteTransaction(const QString& id);
   void deleteTagSplitsList(const QString& txId, const QList<int>& splitIdList);
@@ -411,8 +466,9 @@ private:
   int upgradeToV5();
   int upgradeToV6();
   int upgradeToV7();
+  int upgradeToV8();
 
-  int createTables(int version = std::numeric_limits<int>::max());
+  int createTables();
   void createTable(const MyMoneyDbTable& t, int version = std::numeric_limits<int>::max());
   bool alterTable(const MyMoneyDbTable& t, int fromVersion);
   void clean();
@@ -422,6 +478,18 @@ private:
     return (m_driver->tables(tt, static_cast<const QSqlDatabase&>(*this)));
   };
 
+  /**
+   * @brief Ensure the storagePlugin with iid was setup
+   *
+   * @throws MyMoneyException in case of an error which makes the use
+   * of the plugin unavailable.
+   */
+  bool setupStoragePlugin(QString iid);
+
+  void insertStorableObject(const databaseStoreableObject& obj, const QString& id);
+  void updateStorableObject(const databaseStoreableObject& obj, const QString& id);
+  void deleteStorableObject(const databaseStoreableObject& obj, const QString& id);
+
   // data
   KSharedPtr<MyMoneyDbDriver> m_driver;
 
@@ -448,7 +516,11 @@ private:
   long unsigned m_reports;
   long unsigned m_kvps;
   long unsigned m_budgets;
-  // next id to use (for future archive)
+  long unsigned m_onlineJobs;
+  long unsigned m_payeeIdentifier;
+
+  // Cache for next id to use
+  // value 0 means data is not available and has to be loaded from the database
   long unsigned m_hiIdInstitutions;
   long unsigned m_hiIdPayees;
   long unsigned m_hiIdTags;
@@ -458,6 +530,9 @@ private:
   long unsigned m_hiIdSecurities;
   long unsigned m_hiIdReports;
   long unsigned m_hiIdBudgets;
+  long unsigned m_hiIdOnlineJobs;
+  long unsigned m_hiIdPayeeIdentifier;
+
   // encrypt option - usage TBD
   QString m_encryptData;
 
@@ -517,5 +592,10 @@ private:
     * @sa setStartDate()
     */
   static QDate m_startDate;
+
+  /**
+    *
+    */
+  QSet<QString> m_loadedStoragePlugins;
 };
 #endif // MYMONEYSTORAGESQL_H
